#include "Cache.hpp"

/******************************************************************************

    HawkCreek software

    Notes:
    - order the if() statements to minimize jumps
    - do I really need a doubly linked list instead of a singly linked one?
    - I need an update() function to sync the cached in surface with the
      possibly modified source surface, otherwise the cache might have an older
      version of the surface and keep blitting that

    History of Changes:
    - 20.09.1999: file created by BuschnicK
    - 07.10.1999: started to implement a way to lock/unlock surfaces in the cache,
      meaning that they won't be moved in the cache and they won't be discarded,
      usefull for mousecursors etc, things that will stay on the screen for a long
      time. Note: This doesn't have anything to do with directdraw surface locking/
      unlocking -> choose a better name ?
    - 10.10.1999: documentation added
    - 10.10.1999: removed the locking stuff again - it was too buggy and complicated
      to get right/fast, still want that functionality, though
    - 06.11.1999: added Invalidate(), to allow a cache rebuild after a taskswitch
	- 08.11.2000: reimplememt the priority queue with a data structure more suited 
				  to the task ( STL heap or priority queue )
******************************************************************************/

#include "general\helpmacros.hpp"
#include "BlackFish\dxdraw.hpp"
#include "BlackFish\Surface.hpp"
#include "BlackFish\paint.hpp"

/*
 * Creates the cache by allocating 'NrOfCacheSurfaces' video memory surfaces of size
 * 256*256, it then creates a doubly linked list of tCACHEDRECTNODES which point to
 * the individual tiles in the cachesurfaces. This list is used as a priority queue,
 * i.e. The last used(blitted) tile is always the first element in the list, and the
 * last element is the least recently used one, which gets discarded if a blit is
 * requested and the desired tile isn't already in the cache
 *
 * Note: CacheRectWidth and -Height must be powers of two and must evenly divide
 * a 256*256 CacheSurface
 *
 * !! NrOfCacheSurfaces should be calculated to fill all available videomemory
 */
cCACHE::cCACHE(
    tUINT32 nCacheRectWidth,
    tUINT32 nCacheRectHeight,
    tUINT32 TransparentColor,
    tUINT32 nNrOfCacheSurfaces )
{
    ASSERT( DirectDraw != NULL );
    ASSERT( nNrOfCacheSurfaces > 0 );

    CacheRectWidth          = nCacheRectWidth;
    CacheRectHeight         = nCacheRectHeight;
    NrOfCacheSurfaces       = nNrOfCacheSurfaces;
    tUINT32 RectsPerSurface = ( CacheSurfaceWidth / CacheRectWidth ) * ( CacheSurfaceWidth / CacheRectHeight );
    NrOfCacheRects          = NrOfCacheSurfaces * RectsPerSurface;
    NrOfFreeCacheRects      = NrOfCacheRects;

    // create the surfaces in videomemory
    cPAINT *Painter = NULL;
    Painter = new cPAINT();
    ASSERT( Painter != NULL );

    CacheSurface = NULL;
    CacheSurface = new cSURFACE*[NrOfCacheSurfaces];
    for ( tUINT32 i=0; i<NrOfCacheSurfaces; i++)
    {
        CacheSurface[i] = new cSURFACE();
        CacheSurface[i]->Create( CacheSurfaceWidth, CacheSurfaceWidth, TRUE );  // square
        CacheSurface[i]->SetTransparentColor( TransparentColor );
        Painter->SetDrawingTarget( CacheSurface[i] );
        Painter->Fill( 0 );
    }
    SAFEDELETE( Painter );

    // create the cachesurfacerects priority queue
    First = NULL;
    Last  = NULL;
    tCACHEDRECTNODE *Temp       = NULL;
    tCACHEDRECTNODE *Iterator   = NULL;
    tUINT32 RectsPerRow = CacheSurfaceWidth / CacheRectWidth;
    // for each CacheSurface
    for ( tUINT32 SurfaceNr = 0; SurfaceNr < NrOfCacheSurfaces; SurfaceNr++ )
        // for each Rectangle in that surface
        for ( tUINT32 RectNr = 0; RectNr < RectsPerSurface; RectNr++ )
        {
            tUINT32 RowNr = RectNr / RectsPerRow;

            Temp = new tCACHEDRECTNODE;

            Temp->Prev              = Iterator;
            Temp->Next              = NULL;
            Temp->CacheSurface      = CacheSurface[SurfaceNr];
            Temp->SourceSurface     = NULL;
            Temp->CacheRect.top     = RowNr * CacheRectHeight;
            Temp->CacheRect.left    = ( RectNr - ( RectsPerRow * RowNr )) * CacheRectWidth;
            Temp->CacheRect.bottom  = Temp->CacheRect.top + CacheRectHeight;
            Temp->CacheRect.right   = Temp->CacheRect.left + CacheRectWidth;

            if ( Iterator != NULL )     // special case: first node - iterator = NULL
                Iterator->Next = Temp;
            else if ( Iterator == NULL )
                First = Temp;           // remember first node as list entry point
            Iterator = Temp;
        }
    Last = Temp;                        // remember last node as LRU
}


/*
 * destructor
 */
cCACHE::~cCACHE( )
{
    for ( tUINT32 i=0; i<NrOfCacheSurfaces; i++)
        SAFEDELETE( CacheSurface[i] );
    SAFEDELETE( CacheSurface );
    tCACHEDRECTNODE *Temp = First;
    while ( Temp->Next != NULL )
    {
        Temp = Temp->Next;
        SAFEDELETE( Temp->Prev );
    }
    SAFEDELETE( Temp );
}

/*
 * removes every surface from the cache which causes them to be reloaded from systemmmory
 * usefull to recover from a corrupted cache, after a taskswitch for example
 */
void
cCACHE::Invalidate( void )
{
    NrOfFreeCacheRects = NrOfCacheRects;
    tCACHEDRECTNODE *Temp = First;
    while ( Temp != NULL )
    {
        if ( Temp->SourceSurface != NULL )
            Temp->SourceSurface->SetCacheStatus( NULL );
        Temp->SourceSurface = NULL;
        Temp = Temp->Next;
    }
}

/* 
 *
 */
void
cCACHE::Invalidate( cSURFACE *Surface )
{
	tCACHEDRECTNODE *Node = (tCACHEDRECTNODE *) Surface->GetCacheStatus( );
	if ( Node != NULL )
	{
		NrOfFreeCacheRects++;
		Node->SourceSurface = NULL;
		Surface->SetCacheStatus( NULL );
	}
}

/*
 * Debug function, blits the contents of the specified cachesurface to the
 * destination surface
 */
DEBUGCODE
(
void
cCACHE::Dump( cSURFACE *Dest, tUINT32 SurfaceNr )
{
    RECT DestRect;
	SetRect( &DestRect, 800 - 256, 600 - 256, 800, 600 );
    Dest->Blit( CacheSurface[SurfaceNr], NULL, &DestRect );
}
)

/*
 * Move node 'Node' up in the list, so it becomes the first = last used one
 */
void
cCACHE::MoveNodeUp( tCACHEDRECTNODE *Node )
{
    ASSERT( Node != NULL );

    tCACHEDRECTNODE *TempNode1 = NULL;
    tCACHEDRECTNODE *TempNode2 = NULL;

    if ( Node == First )                // resorting the list, current surface will be the first in the list
    {                                   // current surface already is the first, no need to sort
    }                                   // note: this takes care of the first == last issue, too
    else if ( Node == Last )            // current surface is the last
    {                                   // insert current last node just before the first node and adjust
        TempNode1       = Last->Prev;   // the first and last pointers accordingly
        TempNode1->Next = NULL;
        Last->Next      = First;
        Last->Prev      = NULL;
        First->Prev     = Last;
        First           = Last;
        Last            = TempNode1;
    }
    else                                // surface is neither first nor last in the list
    {                                   // remove current node from the list and insert it before the first
        TempNode1       = Node->Prev;   // adjust first pointer
        TempNode2       = Node->Next;
        TempNode1->Next = TempNode2;
        TempNode2->Prev = TempNode1;
        Node->Prev      = NULL;
        Node->Next      = First;
        First->Prev     = Node;
        First           = Node;
    }
}

/*
 * caches in a new surface
 */
void
cCACHE::CacheInSurface( cSURFACE *Surface )
{
    ASSERT( Surface != NULL );

    tBOOL Found = FALSE;
    tCACHEDRECTNODE *Temp = NULL;

    RECT TempRect;
    Surface->GetSize( &TempRect );

    if ( NrOfFreeCacheRects > 0 )   // we have some free cache space left for the new surface
    {
        NrOfFreeCacheRects--;
        Temp  = First;
        Found = FALSE;
        while ( !Found )            // find a cacherect wich isn't used, yet
        {
            if ( Temp->SourceSurface == NULL )  // no source surface for this rect -> rect is unused
            {
                Found = TRUE;
            } 
			else
            {
                Temp = Temp->Next;  // haven't found a suitable rect, yet -> continue through the list
            }
        }
        // we found a suitable CacheSurface with enough room left for the new surface
        TempRect.top    += Temp->CacheRect.top;
        TempRect.left   += Temp->CacheRect.left;
        TempRect.bottom += Temp->CacheRect.top;
        TempRect.right  += Temp->CacheRect.left;
        Temp->SourceSurface = Surface;  // !! cyclcic dependancy really necessary here?!
        Temp->SourceSurface->SetCacheStatus( (void *)Temp );
        Temp->CacheSurface->Blit( Surface, NULL, &TempRect );  // blit the sourcesurface into the cachesurface
    }
    else
    {
        // no CacheSurface has room left, discard oldest cached-in surface...
        // note: the new sourcesurface will be inserted at the end of the list and later
        // resorted to be the first, !! probably more efficient to insert it directly as the first
        TempRect.top    += Last->CacheRect.top;
        TempRect.left   += Last->CacheRect.left;
        TempRect.bottom += Last->CacheRect.top;
        TempRect.right  += Last->CacheRect.left;
        Last->SourceSurface->SetCacheStatus( NULL );
        Last->SourceSurface = Surface;   // !! cyclcic dependancy really necessary here?!
        Last->SourceSurface->SetCacheStatus( (void *)Last );
        Last->CacheSurface->Blit( Surface, NULL, &TempRect );
    }
}

/*
 * Checks if a surface is already in the cache, if so, it moves up in the priority
 * queue to be the last used one and gets blitted. If the requested surface is
 * not already in the cache it gets blitted from system memory into a cache surface
 * and then the desired blit to the target surface is done from cachesurface to destsurface
 */
void
cCACHE::Blit( cSURFACE *Source, RECT *SourceRect, cSURFACE *Dest, RECT *DestRect )
{
    ASSERT( Source != NULL );
    ASSERT( Dest != NULL );
    // Note: the rects are allowed to be null, meaning the entire surface is to be blitted()

    tCACHEDRECTNODE *Node     = NULL;
    Node = ( tCACHEDRECTNODE * ) Source->GetCacheStatus();

    // the surface is not yet in the cache
    if ( Node == NULL ) CacheInSurface( Source );

    Node = ( tCACHEDRECTNODE * ) Source->GetCacheStatus();  // update node info
    ASSERT( Node != NULL );

    MoveNodeUp( Node );

    RECT CacheSourceRect;
    if ( SourceRect != NULL)
    {
        CacheSourceRect.top     = Node->CacheRect.top  + SourceRect->top;   // !! should check for legal values
        CacheSourceRect.left    = Node->CacheRect.left + SourceRect->left;
        CacheSourceRect.bottom  = Node->CacheRect.top  + SourceRect->bottom;
        CacheSourceRect.right   = Node->CacheRect.left + SourceRect->right;
    } 
	else                              // the sourcerect is NULL, meaning that the user wants to blit
    {                                   // the entire surface
        CacheSourceRect.top     = Node->CacheRect.top;
        CacheSourceRect.left    = Node->CacheRect.left;
        CacheSourceRect.bottom  = Node->CacheRect.bottom;
        CacheSourceRect.right   = Node->CacheRect.right;
    }
    ASSERT( Node->SourceSurface == Source );
    Dest->Blit( Node->CacheSurface,  &CacheSourceRect, DestRect );  // finally: blit from cache
}

/*
 * this function is (and should be) an exact copy of cCACHE::Blit() except for the last
 * line, wich calls cSURFACE::Blit() or cSURFACE::BlitTrans() respectively
 * Possible ways to minimize code dublication:
 * - move common code into an extra function PrepareBlit() | I didn't want the overhead
 *   of an additional function call in an often called/important function such as this
 *   - 08.10.1999 did just this, see the functions MoveNodeUp() and CacheInSurface()
 * - Pass a flag to the function, telling it to either blittrans() or blit(), branching
 *   in an if() statement | I didn't want to break consistency with the current cSURFACE
 *   interface of having two seperate functions
 */
void
cCACHE::BlitTrans( cSURFACE *Source, RECT *SourceRect, cSURFACE *Dest, RECT *DestRect )
{
    ASSERT( Source != NULL );
    ASSERT( Dest != NULL );
    // Note: the rects are allowed to be null, meaning the entire surface is to be blitted()

    tCACHEDRECTNODE *Node     = NULL;
    Node = ( tCACHEDRECTNODE * ) Source->GetCacheStatus();

    // the surface is not yet in the cache
    if ( Node == NULL )
    {
        CacheInSurface( Source );
        Node = ( tCACHEDRECTNODE * ) Source->GetCacheStatus();  // update node info
        ASSERT( Node != NULL );
    }

    MoveNodeUp( Node );

    RECT CacheSourceRect;
    if ( SourceRect != NULL)
    {
        CacheSourceRect.top     = Node->CacheRect.top  + SourceRect->top;   // !! should check for legal values
        CacheSourceRect.left    = Node->CacheRect.left + SourceRect->left;
        CacheSourceRect.bottom  = Node->CacheRect.top  + SourceRect->bottom;
        CacheSourceRect.right   = Node->CacheRect.left + SourceRect->right;
    } 
	else                                // the sourcerect is NULL, meaning that the user wants to blit
    {                                   // the entire surface
        CacheSourceRect.top     = Node->CacheRect.top;
        CacheSourceRect.left    = Node->CacheRect.left;
        CacheSourceRect.bottom  = Node->CacheRect.bottom;
        CacheSourceRect.right   = Node->CacheRect.right;
    }
    ASSERT( Node->SourceSurface == Source );
    Dest->BlitTrans( Node->CacheSurface, &CacheSourceRect, DestRect );  // finally: blit from cache
}
