#include "Game.hpp"

/******************************************************************************

	HawkCreek software

	Notes:

	History of Changes:
 ******************************************************************************/

#include <stdlib.h>                     // rand, itoa
#include <mmsystem.h>                   // timeGetTime()

#include "general\debug.hpp"
#include "general\helpmacros.hpp"
#include "general\type.hpp"
#include "BlackFish\BlackFish.hpp"
#include "BlackFish\Cache.hpp"
#include "BlackFish\DXDraw.hpp"
#include "BlackFish\Input.hpp"
#include "BlackFish\Surface.hpp"
#include "BlackFish\PixelConverter.hpp"
#include "BlackFish\paint.hpp"
#include "BlackFish\TileSet.hpp"
#include "BlackFish\Time.hpp"
#include "BlackFish\write.hpp"
#include "BlackFish\sprite.hpp"
#include "Map\Map.hpp"

//////////////////////////////////////////////

/*
 * base class, all other handlers must be derived from this
 * default constructor is initialized with the next handler in the scene
 */
cEVENTHANDLER::cEVENTHANDLER( cEVENTHANDLER *NextHandler = NULL )
	: NextHandler( NextHandler )
{
}

/*
 *
 */
cEVENTHANDLER::~cEVENTHANDLER( )
{
}

/*
 * base handler can't handle any events and just passes them along
 * this method is intended to be overwritten
 */
void
cEVENTHANDLER::HandleEvent( cEVENT *Event )
{
	// default behaviour: can't deal with this event, just pass it along
	// note: if NextHandler is NULL, the Event drops off the end of the chain unhandled
	if ( NextHandler )
		NextHandler->HandleEvent( Event );
}

/*
 * handles all events relating to the map, i.e. mouseclicks in the mapwindow, 
 * mousemovement (scrolling) etc
 */
cMAPEVENTHANDLER::cMAPEVENTHANDLER( 
	cBLACKFISH *BlackFish, 
	cEVENTHANDLER *NextHandler )
	: cEVENTHANDLER( NextHandler ),
 	  BlackFish( BlackFish ),
	  SelectedMarine( NULL ),
	  OldX( 0 ),
	  OldY( 0 )
{
}

/*
 * handles:
 * - left clicks on map-window
 * - mouse-movement at the screen-borders (scrolling)
 */
void
cMAPEVENTHANDLER::HandleEvent( cEVENT *Event )
{
	/**********************************/
	// try MouseClickEvent
	cMOUSECLICKEVENT *MouseClickEvent = dynamic_cast<cMOUSECLICKEVENT *> ( Event );
	if ( MouseClickEvent )	
	{
		tUINT32 ArrayX, ArrayY;
		// leftclick
		if (( MouseClickEvent->LastClick & cMOUSE::Button1 )
		   && ( BlackFish->GetCellForScreen( &ArrayX, &ArrayY, MouseClickEvent->PositionX, MouseClickEvent->PositionY )))
		{
			std::list<cSPRITE *> *List = BlackFish->GetMap( )->GetSpriteList( ArrayX, ArrayY );
			std::list<cSPRITE *>::iterator Iterator;
			for ( Iterator = List->begin( ); Iterator != List->end( ); Iterator++ )
			{
				cMARINESPRITE *Marine = dynamic_cast<cMARINESPRITE *> ( *Iterator );
				if ( Marine != NULL )
				{
					SelectedMarine = Marine;
					if ( Event->Persistant( )) cEVENTHANDLER::HandleEvent( Event );
					return;
				}
				else
				{
				}
			}
			if ( SelectedMarine )
				SelectedMarine->Move( cVECTOR( ArrayX * 32.0, ArrayY * 32.0, 0.0 ));
		}
		// rightclick
		else if ( MouseClickEvent->LastClick & cMOUSE::Button2 )
		{
			SelectedMarine = NULL;
		}

		if ( Event->Persistant( )) cEVENTHANDLER::HandleEvent( Event );
		return;
	}	

	/**********************************/
	// try MouseMoveEvent
	cMOUSEMOVEEVENT *MouseMoveEvent = dynamic_cast<cMOUSEMOVEEVENT*> ( Event );
	if ( MouseMoveEvent )
	{
		BlackFish->Scroll( MouseMoveEvent->PositionX, MouseMoveEvent->PositionY );

		tUINT32 ArrayX, ArrayY;

		if ( BlackFish->GetCellForScreen( &ArrayX, &ArrayY, MouseMoveEvent->PositionX, MouseMoveEvent->PositionY )
		  && (( ArrayX != OldX ) || ( ArrayY != OldY )))
		{
			// highlight the mapcell the mouse cursor hovers above
			BlackFish->GetMap( )->RemoveTopLayer( OldX, OldY );
			OldX = ArrayX;
			OldY = ArrayY;
			BlackFish->GetMap( )->AddTopLayer( ArrayX, ArrayY, 0, 3 );
		}

		if ( Event->Persistant( )) cEVENTHANDLER::HandleEvent( Event );
		return;
	}

	cEVENTHANDLER::HandleEvent( Event );
}

/*
 * handles general keystrokes
 */
cMAINKEYEVENTHANDLER::cMAINKEYEVENTHANDLER(
	HWND Window,
	cEVENTHANDLER *NextHandler )
	: cEVENTHANDLER( NextHandler ),
	  Window( Window )
{
}

/*
 * handles:
 * - ESC key: quit program
 */
void
cMAINKEYEVENTHANDLER::HandleEvent( cEVENT *Event )
{
	/**********************************/
	// try KeyESCEvent
	cKEY_ESC_EVENT *KeyESCEvent = dynamic_cast<cKEY_ESC_EVENT*> ( Event );
	if ( KeyESCEvent )
	{
		PostMessage( Window, WM_CLOSE, 0, 0 );
		if ( Event->Persistant( )) cEVENTHANDLER::HandleEvent( Event );
		return;
	}

	cEVENTHANDLER::HandleEvent( Event );
}

//////////////////////////////////////////////

/*
 * draws miscellaneous info such as framerate, current position etc...
 */
void
cGAME::ShowDebugInfo( void )
{
	static tDOUBLE Frequency = Timer.GetFrequency();
    static tDOUBLE FrameTime, OldFrameTime, FrameRate;
    static tDOUBLE Frames;

    Frames++;
    FrameTime = Timer.GetTime();
    if ( (FrameTime - OldFrameTime) >= Frequency )
    {
        FrameRate = (Frames * Frequency) / (FrameTime - OldFrameTime);
        OldFrameTime = Timer.GetTime();
        Frames = 0;
    }

    char MiscString[20];
    itoa((tUINT32) FrameRate , MiscString, 10 );
    Writer->Out( 3, 0, "FPS:" );
    Writer->Out( 56, 0, MiscString );

    itoa( BlackFish->WorldX(), MiscString, 10);
    Writer->Out( 3,16, "WrldX:" );
    Writer->Out( 56,16, MiscString );

    itoa( BlackFish->WorldY(), MiscString, 10);
    Writer->Out( 3,32, "WrldY:" );
    Writer->Out( 56,32, MiscString );

	cVECTOR Pos;
	Marine[0]->GetPosition( &Pos );
    itoa( Pos.x, MiscString, 10);
    Writer->Out( 3,48, "SprtX:" );
    Writer->Out( 56,48, MiscString );
    itoa( Pos.y, MiscString, 10);
    Writer->Out( 3,64, "SprtY:" );
    Writer->Out( 56,64, MiscString );

	tSINT32 mx, my,w;
	tUINT32 x = 23, y = 42,b;
	Mouse->GetState( &mx, &my, &w, &b );
	BlackFish->GetCellForScreen( &x, &y, mx, my );
    itoa( x, MiscString, 10);
    Writer->Out( 3,80, "CellX:" );
    Writer->Out( 56,80, MiscString );
    itoa( y, MiscString, 10);
    Writer->Out( 3,96, "CellY:" );
    Writer->Out( 56,96, MiscString );
}

/*
 * master setup, called once from InitApplication
 */
cGAME::cGAME( HWND WindowHandle, HINSTANCE Instance, tBOOL FullScreen )
{
tDOUBLE Count = Timer.GetTime( );

    Input = new cINPUT( Instance );

    Mouse = new cMOUSE( WindowHandle, Input );

	KeyBoard = new cKEYBOARD( WindowHandle, Input );

	DirectDraw = new cDXDRAW( WindowHandle, FullScreen );

    Primary = new cSURFACE( );
    Primary->Create( );

    BackBuffer = new cSURFACE( );

	if ( DirectDraw->IsFullScreen( ))
		BackBuffer->Create( Primary ); // We're fullscreen, create a DX flipping chain
	else
		BackBuffer->Create( 800, 600, TRUE ); // We're windowed, create offcreen vidmem backbuffer

	RenderTarget = BackBuffer;

    Painter = new cPAINT();
    Painter->SetDrawingTarget( RenderTarget );

    cPIXELCONVERTER *PixelConverter = NULL;
    PixelConverter = new cPIXELCONVERTER( RenderTarget );
    ASSERT( PixelConverter != NULL );
    tUINT32 TransparentColor = PixelConverter->RGBToColor( 255, 0, 255 );
    SAFEDELETE( PixelConverter );

    Cache64x128 = new cCACHE( 64,128, TransparentColor, 4 );
    Cache64x32  = new cCACHE( 64, 32, TransparentColor, 4 );
    Cache16x16  = new cCACHE( 16, 16, TransparentColor, 1 );

    BlackFish = new cBLACKFISH( RenderTarget, Cache16x16, Cache64x32, Cache64x128, "graphics\\BlackFish.ini" );

    MouseCursor = new cTILESET( "graphics\\mouse.ini", "[MouseCursor]" );
    MouseCursor->SetCache( Cache16x16 );

    Font1 = new cTILESET( "graphics\\font.ini", "[Font1]" );
    Font1->SetCache( Cache16x16 );

    Writer = new cWRITE( );
	Writer->SetDrawingTarget( RenderTarget );
    Writer->SetFont( Font1 );
    Writer->SetSpacing( 0 );

    srand((tUINT32) Timer.GetTime( ));

	MainKeyEventHandler = new cMAINKEYEVENTHANDLER( DirectDraw->GetWindow( ), NULL );
	MapEventHandler		= new cMAPEVENTHANDLER( BlackFish, MainKeyEventHandler );
	RootEventHandler	= MapEventHandler;

	Marine.push_back( new cMARINESPRITE( 
		BlackFish->GetMap( ), 
		769, 
		cVECTOR( 128.0, 
				 128.0, 
				 0.0 )));

*LOG << "\nGame init time: " << (Timer.GetTime( ) - Count) / Timer.GetFrequency( ) << "\n";
}

/*
 *
 */
cGAME::~cGAME( )
{
	for ( tUINT32 i=0; i<Marine.size( ); i++ )
		SAFEDELETE( Marine[i] );

	SAFEDELETE( MouseCursor );
    SAFEDELETE( Input );
    SAFEDELETE( Mouse );
	SAFEDELETE( KeyBoard );
    SAFEDELETE( Painter );
    SAFEDELETE( BlackFish );
    SAFEDELETE( BackBuffer );
    SAFEDELETE( Primary );
	SAFEDELETE( DirectDraw );
    SAFEDELETE( Font1 );
    SAFEDELETE( Writer );
    SAFEDELETE( Cache16x16 );
    SAFEDELETE( Cache64x32 );
    SAFEDELETE( Cache64x128 );
	SAFEDELETE( MapEventHandler );
	SAFEDELETE( MainKeyEventHandler );
}

/*
 * called each frame, the mainloop
 * generates and dispatches the messages to the various handlers and is responsible for
 * pooling input devices and for displaying the back buffer
 */
void
cGAME::Run( void )
{
	// Input Handling Section
	tSINT32 MouseX, MouseY, MouseWheelPos;
	tUINT32 MouseButtons, LastClick, LastKey;

    Mouse->UpdateState( );
	KeyBoard->UpdateState( );

	Mouse->GetState( &MouseX, &MouseY, &MouseWheelPos, &MouseButtons );
	LastClick = Mouse->GetLastClick( );
	LastKey = KeyBoard->GetLastKey( );

	// !! I'm passing references to automatic objects to the event handlers - dangerous
	//    thing to do, works under the assumption that event handlers don't keep a pointer...
	// mouseclick message
	if ( LastClick )
		RootEventHandler->HandleEvent( &cMOUSECLICKEVENT( MouseX, MouseY, MouseWheelPos, MouseButtons, LastClick ));

	// always generate mousemove messages (should only have few handlers)
	if ( TRUE )
		RootEventHandler->HandleEvent( &cMOUSEMOVEEVENT( MouseX, MouseY ));

	switch ( LastKey )
	{
		case KEY_ESCAPE:
			RootEventHandler->HandleEvent( &cKEY_ESC_EVENT( ));
		break;
	}

	for ( tUINT32 i=0; i<Marine.size( ); i++ )
		Marine[i]->Think( );

    Painter->Fill( 0 );		// !! remove
    BlackFish->Render( );

    RECT Source, Dest, Screen;
	SetRect( &Screen, 0, 0, 800, 600 );
    MouseCursor->GetTileRect( &Source );
	MouseCursor->GetTileRect( &Dest );
	OffsetRect( &Dest, MouseX - 7, MouseY - 7 );
	Clip( Screen, &Source, &Dest );
    MouseCursor->BlitTrans( 0, &Source, RenderTarget, &Dest );

	ShowDebugInfo( );

	if ( DirectDraw->IsFullScreen() )
		Primary->Flip( );
	else
	{
		RECT Screen;
        GetClientRect( DirectDraw->GetWindow( ), &Screen);
        ClientToScreen( DirectDraw->GetWindow( ), (POINT*)&Screen.left);
        ClientToScreen( DirectDraw->GetWindow( ), (POINT*)&Screen.right);
		Primary->Blit( BackBuffer, NULL, &Screen );
	}
}

/*
 * restores after a taskswitch
 */
void
cGAME::RetrieveFocus( void )
{
    Cache16x16->Invalidate();
    Cache64x32->Invalidate();
    Cache64x128->Invalidate();
    *LOG << "Cache Invalidated! \n";
}

