#include "input.hpp"

/******************************************************************************
    HawkCreek software
    Notes:
    History of Changes:
    - 06.10.1999: file created by BuschnicK
    - 09.10.1999: documentation
    - 10.10.1999: renamed cDXMOUSE to cMOUSE
	- 31.03.2000: added cKEYBOARD
				  !! should all input devices have a common parent class? 
				     since all have common data and functionality?
					 (on the other hand, we only have keyboard and mouse, so...)
				  Task Switching (alt-tab) behaves strange now, check keyboard
				  exclusive mode FIXED
	- 07.04.2000: mouse has an updatestate() etc and behaves more like the keyboard
				  now, also implemented a GetLastClick() which returns the last
				  pressed and released button
				  !! make Mouse::GetState behave like KeyBoard::GetState ?
				     i.e. give an id to getstate and it returns a value
					 for that id (wheel, x, y, buttons)
	- 22.10.2000: changed cMOUSE::UpdateState() so that GetLastClick now
	              returns the last button that has been pressed and released
******************************************************************************/

#include "general\debug.hpp"
#include "general\helpmacros.hpp"
#include <dinput.h>

/*
 * Create a DirectInput Object needed to create the DirectInputDevices (Mouse etc) lateron
 */
cINPUT::cINPUT( const HINSTANCE &Instance )
{
    HRESULT ReturnValue;

    DirectInput = NULL;

    // Register with the DirectInput subsystem and get a pointer
    // to a IDirectInput interface we can use.
    ReturnValue = DirectInputCreate( Instance, DIRECTINPUT_VERSION, &DirectInput, NULL );
    ASSERT( ReturnValue == DI_OK );
    ASSERT( DirectInput != NULL );
}

/*
 * destructor
 */
cINPUT::~cINPUT( )
{
    RELEASE( DirectInput );
}

///////////////////////////////////////////

/*
 * initializes the keyboard object and acquires it
 */
cKEYBOARD::cKEYBOARD( const HWND &GameWindow, cINPUT *Input )
{
	ASSERT( GameWindow != NULL );
	ASSERT( Input != NULL );

    HRESULT ReturnValue;
    ReturnValue = Input->GetDirectInputInterface()->CreateDevice( GUID_SysKeyboard, &KeyBoard, NULL );
    ASSERT( ReturnValue == DI_OK )

    ReturnValue = KeyBoard->SetDataFormat( &c_dfDIKeyboard );
    ASSERT( ReturnValue == DI_OK );

    // Set the cooperativity level to let DirectInput know how
    // this device should interact with the system and with other
    // DirectInput applications.
    ReturnValue = KeyBoard->SetCooperativeLevel( GameWindow, DISCL_NONEXCLUSIVE | DISCL_FOREGROUND );
    ASSERT( ReturnValue == DI_OK );

    Acquire();

	memset( Key, 0, sizeof( Key ));
	memset( LastKey, 0, sizeof ( LastKey ));
	PressedKey = 0;
}

/*
 *
 */
cKEYBOARD::~cKEYBOARD( )
{
	KeyBoard->Unacquire();
	RELEASE( KeyBoard );
}

/*
 * has to be called to gain access to the keyboard, only necessary when the
 * application is reactivated after loosing it's focus
 */
void
cKEYBOARD::Acquire( void )
{
    HRESULT ReturnValue = KeyBoard->Acquire();
    ASSERT((( ReturnValue == DI_OK ) || ( ReturnValue == S_FALSE )));  // S_FALSE = already acquired
}

/*
 * should be called every frame, updates the internal variables
 */
void
cKEYBOARD::UpdateState( void )
{
	ASSERT( KeyBoard != NULL );

    HRESULT ReturnValue = DIERR_INPUTLOST;
    
    ReturnValue = KeyBoard->GetDeviceState( sizeof( Key ), &Key );
    if ( ReturnValue == DIERR_INPUTLOST )
    {
        ReturnValue = KeyBoard->Acquire();
        ASSERT( ReturnValue == DI_OK );
*LOG << "Reacquired the keyboard device! \n";
    }
    ASSERT( ReturnValue == DI_OK );
        
	for ( tUINT32 i = 0; i < 256; i++ )
		if ( Key[i] & 0x80 )
			if ( ! ( LastKey[i] & 0x80 ))
				if ( PressedKey == 0 ) PressedKey = (tUINT8) i;
	memcpy( LastKey, Key, sizeof( LastKey ));
}

/*
 * returns the last pressed and released key (usefull for getting individual keystrokes = 
 * typing of text, as opposed to continuous keypresses = scrolling)
 */
tUINT8
cKEYBOARD::GetLastKey( void )
{
	tUINT8 ReturnValue = PressedKey;
	PressedKey = 0;
	return ReturnValue;
}

/*
 * translates a DirectInput KeyCode to the aquivalent ASCII value
 * !! american keyboard layout
 * how to I properly translate this?
 */
char
cKEYBOARD::KeyToASCII( tUINT8 KeyCode, tBOOL Shift )
{
	switch ( KeyCode )
	{
	case KEY_SPACE: return ' ';
	case KEY_MINUS: return '-';
	case KEY_EQUALS: return '=';
	case KEY_BACK: return 13;		// !! BackSpace
	case KEY_RETURN: return '\n';	// !! Return?
	case KEY_TAB: return '\t';
	case KEY_LBRACKET: return '(';
	case KEY_RBRACKET: return ')';
	case KEY_SEMICOLON: return ';';
	case KEY_APOSTROPHE: return '\'';
	case KEY_COMMA: return ',';
	case KEY_PERIOD: return '.';
	case KEY_SLASH: return '/';
	case KEY_MULTIPLY: return '*';
	case KEY_NUMPAD0: return '0';
	case KEY_NUMPAD1: return '1';
	case KEY_NUMPAD2: return '2';
	case KEY_NUMPAD3: return '3';
	case KEY_NUMPAD4: return '4';
	case KEY_NUMPAD5: return '5';
	case KEY_NUMPAD6: return '6';
	case KEY_NUMPAD7: return '7';
	case KEY_NUMPAD8: return '8';
	case KEY_NUMPAD9: return '9';
	case KEY_SUBTRACT: return '-';
	case KEY_ADD: return '+';
	case KEY_DECIMAL: return '.';
	case KEY_NUMPADCOMMA: return ',';
	case KEY_DIVIDE: return '/';		
	case KEY_1: return '1';
	case KEY_2: return '2';
	case KEY_3: return '3';
	case KEY_4: return '4';
	case KEY_5: return '5';
	case KEY_6: return '6';
	case KEY_7: return '7';
	case KEY_8: return '8';
	case KEY_9: return '9';
	case KEY_0: return '0';
	case KEY_A: return Shift ? 'A' : 'a';
	case KEY_B: return Shift ? 'B' : 'b';
	case KEY_C: return Shift ? 'C' : 'c';
	case KEY_D: return Shift ? 'D' : 'd';
	case KEY_E: return Shift ? 'E' : 'e';
	case KEY_F: return Shift ? 'F' : 'f';
	case KEY_G: return Shift ? 'G' : 'g';
	case KEY_H: return Shift ? 'H' : 'h';
	case KEY_I: return Shift ? 'I' : 'i';
	case KEY_J: return Shift ? 'J' : 'j';
	case KEY_K: return Shift ? 'K' : 'k';
	case KEY_L: return Shift ? 'L' : 'l';
	case KEY_M: return Shift ? 'M' : 'm';
	case KEY_N: return Shift ? 'N' : 'n';
	case KEY_O: return Shift ? 'O' : 'o';
	case KEY_P: return Shift ? 'P' : 'p';
	case KEY_Q: return Shift ? 'Q' : 'q';
	case KEY_R: return Shift ? 'R' : 'r';
	case KEY_S: return Shift ? 'S' : 's';
	case KEY_T: return Shift ? 'T' : 't';
	case KEY_U: return Shift ? 'U' : 'u';
	case KEY_V: return Shift ? 'V' : 'v';
	case KEY_W: return Shift ? 'W' : 'w';
	case KEY_X: return Shift ? 'X' : 'x';
	case KEY_Y: return Shift ? 'Y' : 'y';
	case KEY_Z: return Shift ? 'Z' : 'z';
	default: return 0;
	}
}

///////////////////////////////////////////

/*
 * Mouse constructor, parameters are the applictaion's mainwindow and a valid Input object.
 * the application is setup so it has exclusive foreground access to the mouse.
 * initial mouse coordinates are screenwidth/2 and screenheight/2 = 800/2, 600/2.
 * the mouse is acquired, so getstate() calls work directly after calling the constructor
 */
cMOUSE::cMOUSE( const HWND &GameWindow, cINPUT *Input )
{
	ASSERT( GameWindow != NULL );
    ASSERT( Input != NULL );

    HRESULT ReturnValue;
    // Obtain an interface to the system mouse device.
    ReturnValue = Input->GetDirectInputInterface()->CreateDevice( GUID_SysMouse, &Mouse, NULL );
    ASSERT( ReturnValue == DI_OK )

    // Set the data format to "mouse format" - a predefined data format
    ReturnValue = Mouse->SetDataFormat( &c_dfDIMouse );
    ASSERT( ReturnValue == DI_OK );

    // Set the cooperativity level to let DirectInput know how
    // this device should interact with the system and with other
    // DirectInput applications.
    ReturnValue = Mouse->SetCooperativeLevel( GameWindow, DISCL_EXCLUSIVE | DISCL_FOREGROUND );
    ASSERT( ReturnValue == DI_OK );

    Acquire();

    PositionX		= 800 / 2;
    PositionY		= 600 / 2;
	WheelPosition	= 0;
	Buttons			= 0;
	LastButtons		= 0;
	LastClick		= 0;
}

/*
 * Unacquire the mouse if it has been acquired and release the object
 */
cMOUSE::~cMOUSE()
{
    Mouse->Unacquire();
    RELEASE( Mouse );
}

/*
 * updates internal mouse state variables, should be called every frame
 */
void
cMOUSE::UpdateState( void )
{
    HRESULT ReturnValue = DIERR_INPUTLOST;
    DIMOUSESTATE MouseState;      // DirectInput mouse state structure

    // get the input's device state, and put the state in MouseState
    ReturnValue = Mouse->GetDeviceState( sizeof(DIMOUSESTATE), &MouseState );
    if ( ReturnValue == DIERR_INPUTLOST )
    {
        ReturnValue = Mouse->Acquire();
        ASSERT( ReturnValue == DI_OK );
*LOG << "Reacquired the mouse device! \n";

		WheelPosition	= 0;		// after losing focus the mouse is reinitialized at pos 0 0 (dx default)
		Buttons			= 0;		// since I don't want the user to enter the game scrolling
		PositionX		= 800 / 2;	// I restart the mouse at the screen center
		PositionY		= 600 / 2;
		LastClick		= 0;
		LastButtons		= 0;
    }
	else
	{
		ASSERT( ReturnValue == DI_OK ); // device is not lost -> ReturnValue must be DI_OK

		PositionX += MouseState.lX;
		if ( PositionX < 0 ) PositionX = 0;
		else if ( PositionX >= 800 ) PositionX = 799;

	    PositionY += MouseState.lY;
		if ( PositionY < 0 ) PositionY = 0;
		else if ( PositionY >= 600 ) PositionY = 599;

		WheelPosition = MouseState.lZ;

		Buttons = 0;

		if ( MouseState.rgbButtons[0] & 0x80 ) Buttons |= Button1;
		if ( MouseState.rgbButtons[1] & 0x80 ) Buttons |= Button2;
		if ( MouseState.rgbButtons[2] & 0x80 ) Buttons |= Button3;

		if (!( Buttons & Button1 ))
		{
			if ( LastButtons & Button1 )
				if ( LastClick == 0 ) LastClick = Button1;
		} 
		if (!( Buttons & Button2 ))
		{
			if ( LastButtons & Button2 )
				if ( LastClick == 0 ) LastClick = Button2;
		} 
		if (!( Buttons & Button3 ))
		{
			if ( LastButtons & Button3 )
				if ( LastClick == 0 ) LastClick = Button3;
		} 

		LastButtons = Buttons;
	}
}

/*
 * returns the last mousebutton which has been pressed and released
 */
tUINT32
cMOUSE::GetLastClick( void )
{
	tUINT32 ReturnValue = LastClick;
	LastClick = 0;
	return ReturnValue;
}

/*
 * returns the mouse position, clips to the screensize = 800x600
 * assumes that the mouse is already acquired
 */
void
cMOUSE::GetState( tSINT32 *PosX, tSINT32 *PosY, tSINT32 *WheelPos, tUINT32 *Buttons )
{
    ASSERT( PosX != NULL );
    ASSERT( PosY != NULL );
	ASSERT( Buttons != NULL );
	ASSERT( WheelPos != NULL );

	*WheelPos	= this->WheelPosition;
	*Buttons	= this->Buttons;
    *PosX		= this->PositionX;
    *PosY		= this->PositionY;
}

/*
 * has to be called to gain access to the mouse, only necessary when the
 * application is reactivated after loosing it's focus
 */
void
cMOUSE::Acquire( void )
{
    HRESULT ReturnValue = Mouse->Acquire();
    ASSERT((( ReturnValue == DI_OK ) || ( ReturnValue == S_FALSE )));  // S_FALSE = already acquired
}
