#include "paint.hpp"

/******************************************************************************
    HawkCreek software
    Notes:
    History of Changes:
    - 24.07.1999: file created by BuschnicK
    - 02.08.1999: write a constructor which sets a target (convenience)
	- 08.10.2000: provide automatic locking() and unlocking()? performance?
	- 16.10.2000: make the Lock/UnLock requirements automatic? i.e. lock if
				  it has to be and unlock if it doesn't? much safer, but also
				  considerably slower
******************************************************************************/

#include "general\debug.hpp"
#include "general\helpmacros.hpp"
#include "BlackFish\Surface.hpp"
#include <ddraw.h>
#include <fstream>		// for GetBMPDimension

#ifndef __WINDOWS_H__
#define __WINDOWS_H__
  #ifndef WIN32_LEAN_AND_MEAN // !! reset this after including windows.h?
  #define WIN32_LEAN_AND_MEAN
  #endif
#include <windows.h>
#endif

/*
 * initialize, unusable in this state
 */
cPAINT::cPAINT( void )
{
    Destination = NULL;
}

/*
 *
 */
cPAINT::~cPAINT( )
{
}

/*
 * Set a cSURFACE object as the drawing target for subsequent
 * calls to drawing functions like PutPixel() and Fill()
 * returns the previous drawing target, so callers may restore that after they are done
 */
cSURFACE *
cPAINT::SetDrawingTarget( cSURFACE *Dest )
{
    ASSERT( Dest != NULL );              // do we have a valid target?
    ASSERT( !Dest->IsLocked( ));         // don't change drawing target if current target is locked

	cSURFACE *ReturnValue = Destination;
    Destination = Dest;
	return ReturnValue;
}

/*
 * Some drawing functions (PutPixel()) require the target surface to be
 * locked and must be called in between a BeginDrawing()/EndDrawing() pair
 */
void
cPAINT::BeginDrawing( void )
{
    ASSERT( Destination != NULL );

    Destination->Lock();

    SurfaceMem = Destination->GetSurfaceMemPointer();
    Pitch = Destination->GetPitch();
}

/*
 * unlock the surface
 */
void
cPAINT::EndDrawing( void )
{
    ASSERT( Destination != NULL );

    Destination->UnLock();
}

/*
 * Surface must NOT be locked
 *
 * Clears the whole surface to the specified color
 */
void
cPAINT::Fill( tUINT16 Color )
{
    ASSERT( Destination != NULL );
    ASSERT( !Destination->IsLocked( ));

    HRESULT ReturnValue = DD_OK;    // !! DEBUGCODE this?
    DDBLTFX BlitFX;

    BlitFX.dwSize       = sizeof( BlitFX );
    BlitFX.dwFillColor  = (DWORD) Color;

    IDirectDrawSurface7 *Surface;
    Surface = Destination->GetDXSurface();
    ASSERT( Surface != NULL );

    Destination->RestoreAllIfLost();

    ReturnValue = Surface->Blt( NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &BlitFX );

    ASSERT( ReturnValue == DD_OK );
}

/*
 * Surface must NOT be locked
 *
 * Clears the whole surface to the specified color
 */
void
cPAINT::Fill( tUINT32 x1, tUINT32 y1, tUINT32 x2, tUINT32 y2, tUINT16 Color )
{
    ASSERT( Destination != NULL );
    ASSERT( !Destination->IsLocked( ));

    HRESULT ReturnValue = DD_OK;    // !! DEBUGCODE this?
    DDBLTFX BlitFX;

    BlitFX.dwSize       = sizeof( BlitFX );
    BlitFX.dwFillColor  = (DWORD) Color;

    IDirectDrawSurface7 *Surface;
    Surface = Destination->GetDXSurface();
    ASSERT( Surface != NULL );

	RECT Rect;
	SetRect( &Rect, x1, y1, x2, y2 );

    Destination->RestoreAllIfLost();

    ReturnValue = Surface->Blt( &Rect, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &BlitFX );

    ASSERT( ReturnValue == DD_OK );
}

/*
 * surface must NOT be locked
 *
 * Loads the bitmap file specified by filename into the surface, converting
 * the color format if necessary (for example the 24bit Bitmap to 16bit
 * surface colordepth conversion is done automatically)
 *
 * Note: the surface and bitmap size must be identical
 */
void
cPAINT::LoadBMP( const char *FileName )
{
    ASSERT( FileName != NULL );
    ASSERT( Destination != NULL );
    ASSERT( !Destination->IsLocked( ));

    IDirectDrawSurface7 *Surface = Destination->GetDXSurface();

    HRESULT ReturnValue     = DD_OK;
    HDC     ImageDCHandle   = NULL;
    HDC     SurfaceDCHandle = NULL;
    HBITMAP BitmapHandle    = NULL;
    BITMAP  Bitmap;

    BitmapHandle = (HBITMAP) LoadImage( NULL, FileName, IMAGE_BITMAP, 0, 0,
                                        LR_LOADFROMFILE | LR_CREATEDIBSECTION );
	if ( BitmapHandle == NULL )
	{
		*LOG << "Error opening bmp file '" << FileName << "'!\n";
		QUIT;
	}

    ReturnValue = (HRESULT) GetObject( BitmapHandle, sizeof( Bitmap ), &Bitmap );
    ASSERT( ReturnValue != 0 ); // bytes returned

    DDSURFACEDESC2 SurfaceDescriptor;
    ZeroMemory( &SurfaceDescriptor, sizeof( SurfaceDescriptor ) );
    SurfaceDescriptor.dwSize = sizeof( SurfaceDescriptor );
    Surface->GetSurfaceDesc( &SurfaceDescriptor );

    ImageDCHandle = CreateCompatibleDC( NULL );
    SelectObject( ImageDCHandle, BitmapHandle );

    ReturnValue = Surface->GetDC( &SurfaceDCHandle );
    ASSERT( ReturnValue == DD_OK );

    StretchBlt( SurfaceDCHandle,
                0, 0, SurfaceDescriptor.dwWidth, SurfaceDescriptor.dwHeight,
                ImageDCHandle,
                0, 0, Bitmap.bmWidth, Bitmap.bmHeight,
                SRCCOPY );

    Surface->ReleaseDC( SurfaceDCHandle);
    DeleteDC( ImageDCHandle );
    DeleteObject( BitmapHandle );
}

/*
 * surface MUST be locked
 *
 * draws a single pixel of color 'Color' at the specified location
 *
 * Note: No clipping is done, so if the pixel coordinates are illegal,
 * i.e. outside the surface, the result is undefined (=crash)
 */
void
cPAINT::PutPixel( tUINT32 X, tUINT32 Y, tUINT16 Color )
{
    ASSERT( Destination != NULL );
    ASSERT( Destination->IsLocked( ));

    *(SurfaceMem + (Y * Pitch) + X) = Color;
}

/*
 * surface MUST be locked
 *
 * no clipping is done, assumes that the coordinates are valid
 * X2 and Y2 are outside the rectangle!!! stupid but windows compatible
 */
void
cPAINT::DrawRect( tSINT32 X1, tSINT32 Y1, tSINT32 X2, tSINT32 Y2, tUINT16 Color )
{
    ASSERT( Destination->IsLocked( ));

	tUINT16 *Mem = SurfaceMem + (Y1 * Pitch) + X1;
	tSINT32 Dist = ( Y2 - Y1 - 1 ) * Pitch;

	// top and bottom line of rect
	for ( tSINT32 x = X1; x < X2; x++ )
	{
		*Mem = Color;
		*(Mem + Dist) = Color;
		Mem++;
	}

	Mem  = SurfaceMem + (Y1 * Pitch) + X1;
	Dist = X2 - X1 - 1;

	// left and right line of rect
	for (tSINT32 y = Y1; y < Y2; y++)
	{
		*Mem = Color;
		*(Mem + Dist) = Color;
		Mem+= Pitch;
	}
}

/*
 *	opens .bmp File, reads out Width and Height,
 *	assumes that Width is found at Byte 19 - 22
 *  Heigth at 23 - 26
 *  returns TRUE if successful
 */
tBOOL
cPAINT::GetBMPDimensions( const char *FileName,
						 tUINT32 *Width,
						 tUINT32 *Height)
{
	// Variables for the reading out of Width and Heigth,
	// probably not neccessary, could use original Width and Height,
	// but seems safer that way
	tUINT32 ReadWidth, ReadHeight;

	std::ifstream BitmapFile;
	BitmapFile.open( FileName );

	if (BitmapFile.good())
	{
		//set getpointer to pos 19 and get width
		BitmapFile.seekg(18);
		BitmapFile.read( (char*) &ReadWidth, sizeof(ReadWidth) );

		//set getpointer to pos 23 (should already be there, but anyway) and get height
		BitmapFile.seekg(22);
		BitmapFile.read( (char*) &ReadHeight, sizeof(ReadHeight) );

		*Width = ReadWidth;
		*Height = ReadHeight;

		BitmapFile.close();
		return TRUE;
	}
	else
	{
		BitmapFile.close();
		return FALSE;
	}
}
