#include "tileset.hpp"

/******************************************************************************
    HawkCreek software
    Notes:
    History of Changes:
    - 25.07.1999: file created by BuschnicK
    - 02.08.1999: !! need a ReadtUINT32() for transparentcolor!!
        (12.10.1999) obsolete, transparent color is not stored in the ini any more
    - 13.10.1999: added the 'SetCache()' and 'blit()' functions
******************************************************************************/

#include "general\debug.hpp"
#include "general\helpmacros.hpp"
#include "IniFile\IniFile.hpp"
#include "BlackFish\paint.hpp"
#include "BlackFish\PixelConverter.hpp"

/*
 * constructor, creates a tileset according to the specifications given in
 * the 'IniFileName' inifile section 'ResourceName'
 * does not initialize a cache (so default behaviour is uncached blits)
 * note: all tiles in a tileset are assumed to be of equal size
 */
cTILESET::cTILESET( const char *IniFileName, const char *ResourceName )
{
    Cache = NULL;
    cINIFILE *IniFile = NULL;

    IniFile = new cINIFILE( IniFileName );
    ASSERT( IniFile != NULL );

    tBOOL ReturnValue = FALSE;
    char BitmapFileName[80] = "";

    ReturnValue = IniFile->ReadString( ResourceName, "FileName", BitmapFileName ); // read corresponding bitmap filename
    ASSERT( ReturnValue == TRUE );

    ReturnValue = IniFile->ReadNumber( ResourceName, "TileWidth", &TileWidth );
    ASSERT( ReturnValue == TRUE );

    ReturnValue = IniFile->ReadNumber( ResourceName, "TileHeight", &TileHeight );
    ASSERT( ReturnValue == TRUE );

    ReturnValue = IniFile->ReadNumber( ResourceName, "BitmapWidth", &SurfaceWidth );
    ASSERT( ReturnValue == TRUE );

    ReturnValue = IniFile->ReadNumber( ResourceName, "BitmapHeight", &SurfaceHeight );
    ASSERT( ReturnValue == TRUE );

    ReturnValue = IniFile->ReadNumber( ResourceName, "NrOfTiles", &NrOfTiles );
    ASSERT( ReturnValue == TRUE );

    cSURFACE *Bitmap = NULL;
    Bitmap = new cSURFACE();
    ASSERT( Bitmap != NULL );

    Bitmap->Create( SurfaceWidth, SurfaceHeight, FALSE );

    cPIXELCONVERTER PixelConverter( Bitmap );
	tUINT32 TransparentColor = PixelConverter.RGBToColor( 255, 0, 255 );
    Bitmap->SetTransparentColor( TransparentColor );	// !! hardcoded transparent color

    cPAINT *Painter = NULL;
    Painter = new cPAINT();
    ASSERT( Painter != NULL );

    Painter->SetDrawingTarget( Bitmap );
    Painter->LoadBMP( BitmapFileName );

    Tile = NULL;
    Tile = new cSURFACE*[NrOfTiles];
    ASSERT( Tile != NULL );

    tUINT32 TilesPerRow = SurfaceWidth / TileWidth;

    tUINT32 RowNr = 0;
    RECT SourceRect;

    for ( tUINT32 TileNr=0; TileNr<NrOfTiles; TileNr++ )
    {
        RowNr = TileNr / TilesPerRow;

        SourceRect.top      = RowNr * TileHeight;
        SourceRect.bottom   = SourceRect.top + TileHeight;
        SourceRect.left     = ( TileNr - ( TilesPerRow * RowNr )) * TileWidth;
        SourceRect.right    = SourceRect.left + TileWidth;

        Tile[TileNr] = new cSURFACE();
        Tile[TileNr]->Create( TileWidth, TileHeight, FALSE );
        Tile[TileNr]->SetTransparentColor( TransparentColor );
        Tile[TileNr]->Blit( Bitmap, &SourceRect, NULL );
    }

    SAFEDELETE( Bitmap );
    SAFEDELETE( Painter );
    SAFEDELETE( IniFile );
}

/*
 * destructor
 */
cTILESET::~cTILESET()
{
    for ( tUINT32 TileNr=0; TileNr<NrOfTiles; TileNr++ )
        SAFEDELETE( Tile[TileNr] );
    SAFEDELETEARRAY( Tile );
}

/*
 * returns the cSURFACE object corresponding to the given TileNr
 */
cSURFACE *
cTILESET::GetTile( tUINT32 TileNr )
{
    return ( Tile[TileNr] );
}

/*
 * initializes the 'Rectangle' with the tile dimensions
 * note: all tiles in a tileset are assumed to be of equal size
 */
void
cTILESET::GetTileRect( RECT *Rectangle )
{
	ASSERT( Rectangle != NULL );
	SetRect( Rectangle, 0, 0, TileWidth, TileHeight );
}

/*
 * Blits tile 'TileNr' onto the Surface 'Dest' using the regions specified by source and
 * destrect, scaling if neccessary. 
 * uses a cache if one specified by 'SetCache' otherwise blits the surface directly
 * !! confusing, since blit in this case means blitting the calling object
 *    somewhere, whereas in cSURFACE blit means blitting something on
 *    the calling object
 */
void
cTILESET::Blit( tUINT32 TileNr, RECT *SourceRect, cSURFACE *Dest, RECT *DestRect )
{
    ASSERT( Dest != NULL );

    if ( TileNr < NrOfTiles )
	{
		if ( Cache )
	        Cache->Blit( Tile[TileNr], SourceRect, Dest, DestRect );
		else 
			Dest->Blit( Tile[TileNr], SourceRect, DestRect );
	}
    else
	{
        *LOG << "Tried to Blit() an invalid tile in cTILESET::Blit. \n";
	}
}

/*
 * Blits tile 'TileNr' onto the Surface 'Dest' using the regions specified by source and
 * destrect, scaling if neccessary. 
 * uses a cache if one specified by 'SetCache' otherwise blits the surface directly
 * uses the transparent color specified in the constructor (255,0,255)
 * !! confusing, since blit in this case means blitting the calling object
 *    somewhere, whereas in cSURFACE blit means blitting something on
 *    the calling object
 */
void
cTILESET::BlitTrans( tUINT32 TileNr, RECT *SourceRect, cSURFACE *Dest, RECT *DestRect )
{
    ASSERT( Dest != NULL );

    if ( TileNr < NrOfTiles )
	{
		if ( Cache )
			Cache->BlitTrans( Tile[TileNr], SourceRect, Dest, DestRect );
		else 
			Dest->BlitTrans( Tile[TileNr], SourceRect, DestRect );
	}
    else
	{
        *LOG << "Tried to Blit() an invalid tile in cTILESET::BlitTrans. \n";
	}
}
