#include "Map.hpp"
/******************************************************************************
	HawkCreekSoftware

	MapClass, should contain complete Map Data and Functionality

	History of changes:
	19.03.2000: file created by Capella
	08.04.2000: constructor for new map now takes a MapHeader Struct 
				as parameter
	03.04.2000: BuschnicK, added a Sprite pointer to tCELLINFO
	24.06.2000: BuschnicK, major rewrite, changed version to 2
				I'm not entirely happy with the new class interface, cause it's 
				still very much a struct with little data hiding, but it needs
				to be very fast (accessed 1000s of times per second)...
*******************************************************************************/

#include "general\debug.hpp"
#include "general\HelpMacros.hpp"
#include "general\vector.hpp"
#include <fstream>
#include <algorithm>
#include <math.h>

using std::list;
using std::vector;
using std::make_heap;
using std::sort_heap;
using std::pop_heap;
using std::find;

//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

/* 
 * default constructor, make empty map 
 */
cMAP::cMAP( void )
{
	Header.Version	= DATAVERSION;
	Header.Width	= 32;
	Header.Height	= 32;
	memset( Header.Name, ' ', sizeof( Header.Name ));
	memcpy( Header.Name, "default map", 12 ); 
	memset( Header.Author, ' ', sizeof ( Header.Author ));
	memcpy( Header.Author, "system generated", 17 );
	memset( Header.Padding, 0, sizeof ( Header.Padding ));

	FileName = "";
	Cell = NULL;
    Cell = new cBASECELL[Header.Width * Header.Height];
    ASSERT( Cell != NULL );
}

/* 
 * constructor to make empty map of certain width and height 
 */
cMAP::cMAP ( tMAPHEADER MapHeader )
{
	memcpy( &Header, &MapHeader, sizeof( MapHeader ));

	// Padding is not allowed to hold any information, so overwrite anything that
	// might be there - saves the caller from having to this himself
	memset( Header.Padding, 0, sizeof ( Header.Padding ));

	Cell = NULL;
    Cell = new cBASECELL[Header.Width * Header.Height];
    ASSERT( Cell != NULL );

	FileName = "";
}

/* 
 * constructor to make Map from File 
 */
cMAP::cMAP( const char* NewFileName )
{
	Cell = NULL;
	Load( NewFileName );
}

/* 
 * destructor, free array of cells 
 */
cMAP::~cMAP( )
{
	DeleteCells();
}

/*
 *	Delete the entire array of cells, to be called by destructor and
 *  before loading new map
 */
void 
cMAP::DeleteCells( void )
{
	SAFEDELETEARRAY( Cell ); // delete topmost layer (the groundtiles) calls cBASECELL destructor which
							 // deletes all layers
}

/*
 * !! sort sprites by worldcoordinates
 * !! use STL set for this (automatically sorted and allows only unique objects (a sprite
 *    pointer can never exist more than once or we have an error))
 */
void
cMAP::InsertSprite( cVECTOR *Position, cSPRITE *Sprite )
{
	ASSERT( Position != NULL );
	ASSERT( Sprite != NULL );

	GetSpriteList((tUINT32) ( Position->x / 32.0 ),
				  (tUINT32) ( Position->y / 32.0 ))->push_back( Sprite );

	// !! sort by y coordinate
}

/*
 *
 */
void
cMAP::RemoveSprite( cVECTOR *Position, cSPRITE *Sprite )
{
	ASSERT( Position != NULL );
	ASSERT( Sprite != NULL );

	GetSpriteList((tUINT32) ( Position->x / 32.0 ),
				  (tUINT32) ( Position->y / 32.0 ))->remove( Sprite );
}

/*
 * scans the map for unused layers and removes these (unused = tileset 0 tile 0)
 */
void
cMAP::Optimize( void )
{
    cCELLINFO *OldCheckTile;
    cCELLINFO *CheckTile;
	
    for ( tUINT32 j = 0; j < Header.Height; j++ )
        for ( tUINT32 i = 0; i < Header.Width; i++ )
        {
            CheckTile = GetFirstLayer( i, j );
			ASSERT( CheckTile != NULL );

            while ( CheckTile->NextLayer != NULL )
            {
                OldCheckTile = CheckTile;
                CheckTile = CheckTile->NextLayer;
                if (( CheckTile->TileNr == 2 ) && ( CheckTile->TileSet == 0 ))
                {
                    OldCheckTile->NextLayer = CheckTile->NextLayer;
                    SAFEDELETE ( CheckTile );
                    CheckTile = OldCheckTile;
				}
			}
		}
}

/*
 * save the map
 * !! currently doesn't include sprites
 */
void
cMAP::Save( const char *FileName )
{
	ASSERT( FileName != NULL );

	// make output file and open as binary
	std::ofstream MapFile( FileName, std::ios::out|std::ios::binary );
	ASSERT( MapFile.good( ));

	// header is supposed to be 256 bytes
	MapFile.write( (char *)&Header, sizeof( Header ));

	// temporary variable used to write to file
	tUINT32 Data = 0;

	for ( tUINT32 j = 0; j < Header.Height; j++ )
		for ( tUINT32 i = 0; i < Header.Width; i++ )
		{
			// write the base layer's sprite info
			Data = GetFirstLayer( i, j )->TileSet;
			MapFile.write( (char *)&Data, sizeof( Data ));
			Data = GetFirstLayer( i, j )->TileNr;
			MapFile.write( (char *)&Data, sizeof( Data ));
 			Data = GetBaseCell( i, j )->GetFlags( );
			MapFile.write( (char *)&Data, sizeof( Data ));
			Data = GetBaseCell( i, j )->GetWall( ) != NULL; // used as a tBOOL here
			MapFile.write( (char *)&Data, sizeof( Data ));

			// get number of top layers
			tUINT32 NrOfLayers = 0;
			cCELLINFO *Iterator = GetFirstLayer( i, j );
			while ( Iterator->NextLayer != NULL )
			{
				NrOfLayers++;
				Iterator = Iterator->NextLayer;
			}

			// write number of top layers
			Data = NrOfLayers;
			MapFile.write( (char *)&Data, sizeof( Data ));

			Iterator = GetFirstLayer( i, j );
			for ( tUINT32 k = 0; k < NrOfLayers; k++ )
			{
				Iterator = Iterator->NextLayer;
				Data = Iterator->TileSet;
				MapFile.write( (char *)&Data, sizeof( Data ));
				Data = Iterator->TileNr;
				MapFile.write( (char *)&Data, sizeof( Data ));
			}

			if ( GetBaseCell( i, j )->GetWall( ) != NULL )
			{
				Data = GetBaseCell( i, j )->GetWall( )->TileSet;
				MapFile.write( (char *)&Data, sizeof( Data ));
				Data = GetBaseCell( i, j )->GetWall( )->TileNr;
				MapFile.write( (char *)&Data, sizeof( Data ));
			}
		}
	MapFile.close( );
}

/*
 * Load a new map from file (without making new map object)
 * !! currently doesn't support sprites
 */
void
cMAP::Load( const char* NewFileName )
{
	ASSERT( NewFileName != NULL );

	// delete old array of cells:
	DeleteCells( );

	// make output file and open as binary
	std::ifstream MapFile( NewFileName, std::ios::in|std::ios::binary );
	ASSERT( MapFile.good( ));

	// header is supposed to be 256 bytes
	MapFile.read( (char *)&Header, sizeof( Header ));
	if ( Header.Version != DATAVERSION )
	{
		*LOG << "Loading of '" << NewFileName << "' failed, expecting version " << DATAVERSION << ", found " << Header.Version << ".\n";
		QUIT;
	}

    Cell = new cBASECELL[Header.Width * Header.Height];
    ASSERT( Cell != NULL );

	tUINT32 Data1 = 0, Data2 = 0;

	for ( tUINT32 j = 0; j < Header.Height; j++ )
		for ( tUINT32 i = 0; i < Header.Width; i++ )
		{
			// read the base layer's sprite info
			tBOOL WallFlag = FALSE;

			MapFile.read( (char *)&Data1, sizeof( Data1 ));
			MapFile.read( (char *)&Data2, sizeof( Data2 ));
			SetFirstLayer( i, j, Data1, Data2, NULL );
			MapFile.read( (char *)&Data1, sizeof( Data1 ));
			GetBaseCell( i, j )->SetFlags( Data1 );

			MapFile.read( (char *)&Data1, sizeof( Data1 ));
			WallFlag = Data1;

			// get number of top layers
			tUINT32 NrOfLayers = 0;
			MapFile.read( (char *)&Data1, sizeof( Data1 ));
			NrOfLayers = Data1;

			cCELLINFO *Iterator = GetFirstLayer( i, j );
			for ( tUINT32 k = 0; k < NrOfLayers; k++ )
			{
				Iterator->NextLayer = new cCELLINFO( );
				ASSERT( Iterator->NextLayer != NULL );
				Iterator = Iterator->NextLayer;

				MapFile.read( (char *)&Data1, sizeof( Data1 ));
				Iterator->TileSet = Data1;
				MapFile.read( (char *)&Data1, sizeof( Data1 ));
				Iterator->TileNr = Data1;
			}

			if ( WallFlag )
			{
				MapFile.read( (char *)&Data1, sizeof( Data1 ));
				MapFile.read( (char *)&Data2, sizeof( Data2 ));
				SetWall( i, j, Data1, Data2 );
				ASSERT( GetBaseCell( i, j )->GetWall( ) != NULL );			
			}
		}
	MapFile.close( );

	FileName = NewFileName;

*LOG << "\n" << NewFileName << " read successfully: \n";
*LOG << "    Name:      " << Header.Name << "\n";
*LOG << "    Author:    " << Header.Author << "\n";
*LOG << "    Dimension: " << Header.Width << "x" << Header.Height << "\n";
}

/*
 *
 */
tSINT32 
cMAP::GetHeuristic( tSINT32 x1, tSINT32 y1, tSINT32 x2, tSINT32 y2 )
{
	tSINT32 abs1 = x2 - x1 > 0 ? x2 - x1 : x1 - x2;
	tSINT32 abs2 = y2 - y1 > 0 ? y2 - y1 : y1 - y2;
	tSINT32 Result = abs1 < abs2 ? abs1 * COST_DIAGONAL : abs2 * COST_DIAGONAL; // min(|dx|,|dy|)
	Result += abs1 > abs2 ? (abs1 - abs2) * COST_ALIGNED : (abs2 - abs1) * COST_ALIGNED; // ||dx|-|dy||
	return Result;
}

/*
 *
 */
tSINT32 
cMAP::GetCost( tSINT32 x1, tSINT32 y1, tSINT32 x2, tSINT32 y2 )
{
	if (( x1 - x2 ) && ( y1 - y2 )) // diagonal movement
		return COST_DIAGONAL;

	return COST_ALIGNED;
}

/*
 *
 */
void
cMAP::DeleteNodes( vector< cNODE * > &Open, vector< cNODE * > &Closed )
{
	tSINT32 i;
	for ( i = 0; i < Open.size( ); i++ )
		SAFEDELETE( Open[i] );
	Open.clear( );

	for ( i = 0; i < Closed.size( ); i++ )
		SAFEDELETE( Closed[i] );
	Closed.clear( );
}

/*
 *
 */
void 
cMAP::AddNode( vector< cNODE * > &Open, vector< cNODE * > &Closed, cNODE *Node )
{
	std::vector< cNODE * >::iterator i;
	for ( i=Closed.begin( ); i != Closed.end( ); i++ )
	{
		if (( Node->x == (*i)->x ) && ( Node->y == (*i)->y ))
			if ( (*i)->Cost < Node->Cost )
			{
				SAFEDELETE( Node );
				return;
			}
			else
			{
				SAFEDELETE( *i );
				Closed.erase( i );
				break;
			}
	}

	for ( i=Open.begin( ); i != Open.end( ); i++ )
	{
		if (( Node->x == (*i)->x ) && ( Node->y == (*i)->y ))
			if ( (*i)->Cost < Node->Cost )
			{
				SAFEDELETE( Node );
				return;
			}
			else
			{
				SAFEDELETE( *i );
				Open.erase( i );
				break;
			}
	}
		
	Open.push_back( Node );
}

bool greater( const cNODE *a, const cNODE *b ); // forward declaration, used in FindPath to sort cNODES *

/*
 * !! not at all optimized:
 * - find a way to reduce allocations/deallocations
 * - find a structure that gives efficient results for searches with two different search keys,
 *   i.e. the open list must be sorted by coordinate (in addnode) and by heuristic (findpath)
 *
 */
tBOOL 
cMAP::FindPath( cVECTOR PointA, cVECTOR PointB, list<cVECTOR> *Path )
{
	vector< cNODE * > Open;
	vector< cNODE * > Closed;

	tUINT32 Iteration = 0;

	tSINT32 X = (tSINT32)(PointA.x / 32.0);
	tSINT32 Y = (tSINT32)(PointA.y / 32.0);
	tSINT32 GoalX = (tSINT32)(PointB.x / 32.0);
	tSINT32 GoalY = (tSINT32)(PointB.y / 32.0);
	Open.push_back( new cNODE( X, Y, 0, GetHeuristic( X, Y, GoalX, GoalY )));
	make_heap( Open.begin( ), Open.end( ), greater );

	while (( !Open.empty( )) && (++Iteration <= MAXITERATIONS))
	{
		pop_heap( Open.begin( ), Open.end( ), greater ); // restructure heap, so smallest element is last in the vector

		cNODE *Current = Open[Open.size() - 1];
		
		Open.pop_back( );	// remove the last element (the smallest) from the vector

		tSINT32 X = Current->x;
		tSINT32 Y = Current->y;

		if (( X == GoalX ) && ( Y == GoalY ))	// we arrived at the goal node
		{
			cNODE *Iterator = Current;
			do	// construct path
			{
				Path->push_front( cVECTOR( Iterator->x * 32.0, Iterator->y * 32.0, 0.0 ));
				Iterator = Iterator->Parent;
			}
			while ( Iterator != NULL );

			SAFEDELETE( Current );
			DeleteNodes( Open, Closed );
			return TRUE;
		}

		tSINT32 NewX, NewY;			
		NewX = X-1; NewY = Y-1;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X; NewY = Y-1;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X+1; NewY = Y-1;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X+1; NewY = Y;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X+1; NewY = Y+1;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X; NewY = Y+1;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X-1; NewY = Y+1;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		NewX = X-1; NewY = Y;
		if ( IsValidCell( NewX, NewY ) && (!IsBlocking( X, Y, NewX, NewY )))
			AddNode( Open, Closed, new cNODE( NewX, NewY, 
				Current->Cost + GetCost( X, Y, NewX, NewY ), 
				GetHeuristic( NewX, NewY, GoalX, GoalY ), Current ));

		make_heap( Open.begin( ), Open.end( ), greater );
		Closed.push_back( Current );
	}

	DeleteNodes( Open, Closed );
	return FALSE;
}

void
cMAP::SetName(const char* MapName)
{
	memcpy(Header.Name, MapName, 64);
}

void
cMAP::SetAuthor(const char* Author)
{
	memcpy(Header.Author, Author, 64);
}

const char*
cMAP::GetName()
{
	return Header.Name;
}

const char*
cMAP::GetAuthor()
{
	return Header.Author;
}
//////////////////////////////////////////////////////////////////////
bool operator > ( const cNODE &a, const cNODE &b )	// object comparison
{
	// returns true if 'a' has a better 'f' (=cost to reach the node + heuristic) than 'b'
	// if nodes have equal 'f' returns true if 'a' has a better heuristic than 'b'
	// this makes the A* algorithm always chase down the nodes, which are already closest
	// to the goal first
	return a.Total > b.Total ? true : ( a.Total == b.Total ? a.Heuristic > b.Heuristic : false );
}

bool greater( const cNODE *a, const cNODE *b )		// pointer comparison
{
	return *a > *b;
}
//////////////////////////////////////////////////////////////////////
/*
 *
 */
cBASECELL::cBASECELL( )
	: FirstLayer( ), 
	  Wall( NULL ), 
	  Flags( 0 ),
	  SpriteList( )
{
}

/*
 *
 */
cBASECELL::~cBASECELL( )
{
	DeleteAllLayers( );
	SAFEDELETE( Wall );
}
